from __future__ import annotations

import os
from abc import ABC
from pathlib import Path
from typing import Sequence

from fs_utils import normalize_path
from utils import HashState


class Context(ABC):
    def __init__(self, repo_root: str, cwd: str):
        self.repo_root = repo_root
        self.cwd = os.path.abspath(cwd)

    def absolute(self, path: str):
        return normalize_path(self.repo_root, self.cwd, path)

    def relative(self, path: str):
        return str(
            os.path.relpath(
                Path(self.repo_root).joinpath(self.absolute(path)), self.cwd
            )
        )

    def sh(self, cmd: str):
        raise NotImplementedError

    def add_dep(self, dep: str):
        self.add_deps([dep])

    def add_deps(self, deps: Sequence[str]):
        raise NotImplementedError

    def input(self, *, file: str, sh: str):
        raise NotImplementedError


class MemorizeContext(Context):
    def __init__(self, repo_root: str, cwd: str, hashstate: HashState):
        super().__init__(repo_root, cwd)
        self.hashstate = hashstate
        self.inputs = []

    def sh(self, cmd: str):
        self.hashstate.record("sh", cmd)

    def add_deps(self, deps: Sequence[str]):
        self.hashstate.record("add_deps", deps)
        for dep in deps:
            self.inputs.append(self.absolute(dep))

    def input(self, *, file: str, sh: str):
        self.hashstate.record("input", file, sh)
        if file is not None:
            self.inputs.append(self.absolute(file))



import hashlib
from typing import Callable, Union

from cache import make_cache_fetcher
from context import MemorizeContext
from loader import Rule
from monitoring import log
from utils import CacheMiss, HashState, MissingDependency
from build_state import BuildState


class PreviewContext(MemorizeContext):
    def __init__(
        self,
        repo_root: str,
        cwd: str,
        hashstate: HashState,
        dep_fetcher: Callable[[str], str],
        cache_fetcher: Callable[[str, str], str],
    ):
        super().__init__(repo_root, cwd, hashstate)

        self.dep_fetcher = dep_fetcher
        self.cache_fetcher = cache_fetcher

    def input(self, *, file: str = None, sh: str = None):
        super().input(file=file, sh=sh)
        if file is not None:
            return self.dep_fetcher(self.absolute(file))
        else:
            return self.cache_fetcher(
                self.hashstate.state(),
                hashlib.md5(sh.encode("utf-8")).hexdigest(),
            )


def make_dep_fetcher(build_state: BuildState):
    def dep_fetcher(input_path, *, flags="r") -> Union[str, bytes]:
        try:
            if input_path in build_state.target_rule_lookup:
                # this input may be stale / unbuilt
                # if so, do not read it, but instead throw MissingDependency
                if build_state.target_rule_lookup[input_path] not in build_state.ready:
                    raise MissingDependency(input_path)
                # so it's already ready for use!

            with open(input_path, flags) as f:
                return f.read()
        except FileNotFoundError:
            raise MissingDependency(input_path)

    return dep_fetcher


def get_deps(build_state: BuildState, rule: Rule):
    """
    Use static dependencies and caches to try and identify as *many*
    needed dependencies as possible, without *any* spurious dependencies.
    """
    hashstate = HashState()
    cache_fetcher = make_cache_fetcher(build_state.cache_directory)
    dep_fetcher = make_dep_fetcher(build_state)

    log(f"Looking for static dependencies of {rule}")
    for dep in rule.deps:
        if dep in build_state.target_rule_lookup:
            if build_state.target_rule_lookup[dep] not in build_state.ready:
                log(
                    f"Static dependency {build_state.target_rule_lookup[dep]} of {rule} is not ready, skipping impl"
                )
                # static deps are not yet ready
                break
        hashstate.update(dep.encode("utf-8"))
        try:
            dep_fetcher(dep, flags="rb")
            with open(dep, "rb") as f:
                hashstate.update(f.read())
        except MissingDependency:
            # get static deps before running the impl!
            # this means that a source file is *missing*, but the error will be thrown in enqueue_deps
            break
    else:
        ctx = PreviewContext(
            build_state.repo_root,
            rule.location,
            hashstate,
            dep_fetcher,
            cache_fetcher,
        )
        ok = False
        try:
            log(f"Running impl of {rule} to discover dynamic dependencies")
            rule.impl(ctx)
            log(f"Impl of {rule} completed with discovered deps: {ctx.inputs}")
            ok = True
        except CacheMiss:
            log(f"Cache miss while running impl of {rule}")
            pass  # stops context execution
        except MissingDependency as e:
            log(f"Dependencies {e.paths} were unavailable while running impl of {rule}")
            pass  # dep already added to ctx.inputs
        # if `ok`, hash loaded dynamic dependencies
        if ok:
            log(
                f"Runtime dependencies resolved for {rule}, now checking dynamic dependencies"
            )
            for input_path in ctx.inputs:
                hashstate.update(input_path.encode("utf-8"))
                try:
                    data = dep_fetcher(input_path, flags="rb")
                except MissingDependency as e:
                    # this dependency was not needed for deps calculation
                    # but is not verified to be up-to-date
                    ok = False
                    log(
                        f"Dynamic dependencies {e.paths} were not needed for the impl, but are not up to date"
                    )
                    break
                else:
                    hashstate.update(data)
        return (
            hashstate.state() if ok else None,
            ctx.inputs + rule.deps,
        )
    return None, rule.deps



import sys
import pathlib
from typing import Callable, Optional, Sequence, TYPE_CHECKING, Union

sys.path.append(str(pathlib.Path(__file__).parent.absolute()))


if TYPE_CHECKING:
    """
    Below are some stubs to make imports in BUILD and rules.py files more pleasant
    see loader.py for injected implementation
    """

    def load(path: str):
        ...

    def callback(
        *,
        name: Optional[str] = None,
        deps: Sequence[str] = (),
        impl: Callable,
        out: Union[str, Sequence[str]] = (),
    ):
        ...

    def find(path: str):
        ...



from pathlib import Path
from shutil import rmtree
from subprocess import CalledProcessError

from cache import get_cache_output_paths
from execution import build
from fs_utils import copy_helper
from monitoring import log
from preview_execution import get_deps
from utils import BuildException, MissingDependency
from work_queue import enqueue_deps
from build_state import BuildState


def worker(build_state: BuildState, index: int):
    scratch_path = Path(build_state.repo_root).joinpath(Path(f".scratch_{index}"))
    if scratch_path.exists():
        rmtree(scratch_path)

    while True:
        todo = build_state.work_queue.get()
        if todo is None:
            return

        try:
            build_state.status_monitor.update(index, "Parsing: " + str(todo))

            log(f"Target {todo} popped from queue by worker {index}")

            # only from caches, will never run a subprocess
            cache_key, deps = get_deps(build_state, todo)

            if cache_key is None:
                # unable to compute cache_key, potentially because not all deps are ready
                log(
                    f"Target {todo} either has unbuilt dependencies, "
                    f"or does not have a cached dynamic dependency resolved"
                )
                deps_ready = not enqueue_deps(build_state, todo, deps)
                if deps_ready:
                    log("Apparently it is missing an input cache in the impl")
                else:
                    log("Apparently it is waiting on unbuilt dependencies")
            else:
                log(f"All the dependencies of target {todo} are ready: {deps}")
                # if the cache_key is ready, *all* the deps must be ready, not just the discoverable deps!
                deps_ready = True

            if deps_ready:
                done = False
                # first check if we're already cached!
                if cache_key:
                    cache_location, cache_output_names = get_cache_output_paths(
                        build_state.cache_directory, todo, cache_key
                    )
                    if cache_location.exists():
                        log(f"Target {todo} is in the cache")
                        try:
                            copy_helper(
                                src_root=cache_location,
                                src_names=cache_output_names,
                                dest_root=build_state.repo_root,
                                dest_names=todo.outputs,
                            )
                        except FileNotFoundError:
                            raise BuildException(
                                "Cache corrupted. This should never happen unless you modified the cache "
                                "directory manually! If so, delete the cache directory and try again."
                            )
                        else:
                            done = True

                if not done:
                    # time to execute! but *not* inside the lock
                    # when we release the lock, stuff may change outside, but
                    # we don't care since *our* dependencies (so far) are all available
                    log(f"Target {todo} is not in the cache, rerunning...")
                    build_state.status_monitor.update(index, "Building: " + str(todo))
                    try:
                        # if cache_key is None, we haven't finished evaluating the impl, so we
                        # don't know all the dependencies it could need. Therefore, we must
                        # run it in the working directory, so the impl can find the dependencies it needs
                        # Then, we run it *again*, to verify that the dependencies are accurate
                        in_sandbox = cache_key is not None

                        if not in_sandbox:
                            log(
                                f"We don't know the dependencies of {todo}, "
                                f"so we are running the impl in the root directory to find out!"
                            )
                            cache_key = build(
                                build_state, todo, deps, scratch_path=None
                            )
                            # now, if no exception has thrown, all the deps are available to the deps finder
                            alt_cache_key, deps = get_deps(build_state, todo)
                            assert (
                                cache_key == alt_cache_key
                            ), "An internal error has occurred"
                            try:
                                alt_cache_key = build(
                                    build_state,
                                    todo,
                                    deps,
                                    scratch_path=scratch_path,
                                )
                            except MissingDependency:
                                raise BuildException("An internal error has occurred.")
                            except CalledProcessError as e:
                                build_state.status_monitor.stop()
                                print(e.cmd)
                                print(e)
                                raise BuildException(
                                    f"The dependencies for target {todo} are not fully specified, "
                                    f"as it failed to build when provided only with them."
                                )
                            assert (
                                cache_key == alt_cache_key
                            ), "An internal error has occurred"
                        else:
                            log(
                                f"We know all the dependencies of {todo}, so we can run it in a sandbox"
                            )
                            build(build_state, todo, deps, scratch_path=scratch_path)
                        cache_location, cache_output_names = get_cache_output_paths(
                            build_state.cache_directory, todo, cache_key
                        )

                        log(
                            f"Target {todo} has been built fully! Caching to {cache_location}"
                        )
                        copy_helper(
                            src_root=scratch_path,
                            dest_root=cache_location,
                            src_names=todo.outputs,
                            dest_names=cache_output_names,
                        )

                        done = True
                    except MissingDependency as d:
                        log(
                            f"Target {todo} failed to fully build because of the missing dynamic "
                            f"dependencies: {d.paths}, requeuing"
                        )
                        scheduled = enqueue_deps(build_state, todo, d.paths)
                        if not scheduled:
                            # due to race conditions, our dependencies are actually all ready now!
                            # no one else will enqueue us, so it is safe to enqueue ourselves
                            build_state.work_queue.put(todo)
                            build_state.status_monitor.move(total=1)

                    if scratch_path.exists():
                        rmtree(scratch_path)

                if done:
                    with build_state.scheduling_lock:
                        build_state.ready.add(todo)
                        # no one will ever add us back, since we are in `ready`
                        build_state.scheduled_but_not_ready.remove(todo)
                        # now it's time to set up our dependents
                        # we need to be inside the lock even if we have no dependents, in case
                        # we *gain* dependents from another thread which could have held the lock!
                        for dependent in todo.runtime_dependents:
                            dependent.pending_rule_dependencies.remove(todo)
                            if not dependent.pending_rule_dependencies:
                                # this guy is ready to go
                                build_state.work_queue.put(dependent)
                                build_state.status_monitor.move(total=1)

            # either way, we're done with this task for now
            build_state.status_monitor.move(curr=1)
            build_state.work_queue.task_done()
        except Exception:
            build_state.status_monitor.stop()
            raise



from threading import Thread

from build_state import BuildState
from build_worker import worker
from monitoring import create_status_monitor
from utils import BuildException


def run_build(build_state: BuildState, target: str, num_threads: int):
    if target not in build_state.target_rule_lookup:
        raise BuildException(f"Target {target} not found in BUILD files.")
    root_rule = build_state.target_rule_lookup[target]

    build_state.status_monitor = create_status_monitor(num_threads)

    build_state.scheduled_but_not_ready.add(root_rule)
    build_state.work_queue.put(root_rule)
    build_state.status_monitor.move(total=1)

    thread_instances = []
    for i in range(num_threads):
        thread = Thread(
            target=worker,
            args=(build_state, i),
        )
        thread_instances.append(thread)
        thread.start()
    build_state.work_queue.join()
    for _ in range(num_threads):
        build_state.work_queue.put(None)
    for thread in thread_instances:
        thread.join()

    build_state.status_monitor.stop()

    if build_state.scheduled_but_not_ready:
        # there is a dependency cycle somewhere!
        base = root_rule
        chain = []
        pos = base
        while True:
            if pos in chain:
                chain.append(pos)
                raise BuildException(
                    f"Circular dependency detected: Rule {pos} depends on itself "
                    f"through the path: {' -> '.join(map(str, chain))}"
                )
            else:
                chain.append(pos)
                pos = pos.pending_rule_dependencies[0]



from __future__ import annotations

import os

import click

from build_coordinator import run_build
from build_state import BuildState
from fs_utils import find_root, get_repo_files
from loader import load_rules


@click.command()
@click.argument("target")
@click.option("--threads", "-t", default=8)
@click.option("--cache-directory", default=".cache")
def cli(target: str, threads: int, cache_directory: str):
    """
    This is a `make` alternative with a simpler syntax and some useful features.
    """
    repo_root = find_root()
    os.chdir(repo_root)

    target_rule_lookup = load_rules()
    all_files = get_repo_files()
    source_files = set(all_files) - target_rule_lookup.keys()

    run_build(
        BuildState(
            target_rule_lookup=target_rule_lookup,
            source_files=source_files,
            cache_directory=cache_directory,
            repo_root=repo_root,
        ),
        target,
        threads,
    )


if __name__ == "__main__":
    cli()



from __future__ import annotations

import os
import sys
from dataclasses import dataclass, field
from glob import glob
from typing import Callable, Dict, List, Optional, Sequence, Union

from utils import BuildException
from fs_utils import find_root, get_repo_files, normalize_path

from context import Context


@dataclass(eq=False)
class Rule:
    name: Optional[str]
    location: str
    deps: Sequence[str]
    impl: Callable[[Context], None]
    outputs: Sequence[str]

    runtime_dependents: List[Rule] = field(default_factory=list)
    pending_rule_dependencies: List[Rule] = field(default_factory=list)

    def __hash__(self):
        return hash(id(self))

    def __str__(self):
        if self.name:
            return self.name
        elif len(self.outputs) == 1:
            return self.outputs[0]
        else:
            return f"<anonymous rule from {self.location}/BUILD>"


def make_callback(
    repo_root: str,
    build_root: str,
    # output parameter
    target_rule_lookup: Dict[str, Rule],
):
    def add_target_rule(target, rule):
        if target in target_rule_lookup:
            raise BuildException(
                f"The target `{target}` is built by multiple rules. Targets can only be produced by a single rule."
            )
        target_rule_lookup[target] = rule

    def callback(
        *,
        name: Optional[str] = None,
        deps: Sequence[str] = (),
        impl: Callable,
        out: Union[str, Sequence[str]] = (),
    ):
        if isinstance(out, str):
            out = [out]
        rule = Rule(
            name=name,
            location=build_root,
            deps=[normalize_path(repo_root, build_root, dep) for dep in deps],
            impl=impl,
            outputs=[normalize_path(repo_root, build_root, output) for output in out],
        )
        for output in rule.outputs:
            add_target_rule(output, rule)

        if name is not None:
            add_target_rule(name, rule)

    def find(path):
        return [
            os.path.relpath(path, build_root)
            for path in glob(
                normalize_path(repo_root, build_root, path),
                recursive=True,
            )
        ]

    return callback, find


class Struct:
    def __init__(self, entries):
        self.__dict__.update(entries)


def make_load_rules(repo_root: str, rules_root: str):
    def load_rules(path):
        path = normalize_path(repo_root, rules_root, path)
        if not path.endswith(".py"):
            raise BuildException(f"Cannot import from a non .py file: {path}")

        __builtins__["load"] = make_load_rules(repo_root, path)
        cached_callback = __builtins__["callback"]
        cached_find = __builtins__["find"]
        del __builtins__["callback"]
        del __builtins__["find"]
        # We hide the callback here, since you should not be running the
        # callback (or anything else!) in an import, but just providing defs
        frame = {"__builtins__": __builtins__}
        reset_mock_imports(frame, ["load"])
        with open(path) as f:
            exec(f.read(), frame)
        __builtins__["callback"] = cached_callback
        __builtins__["find"] = cached_find
        return Struct(frame)

    return load_rules


def reset_mock_imports(frame, targets):
    exec(
        "import buildtool; "
        + " ".join(f"buildtool.{target} = {target};" for target in targets),
        frame,
    )


def load_rules():
    repo_root = find_root()
    src_files = get_repo_files()
    build_files = [file for file in src_files if file.split("/")[-1] == "BUILD"]
    target_rule_lookup: Dict[str, Rule] = {}
    frame = {}
    sys.path.insert(0, repo_root)
    for build_file in build_files:
        with open(build_file) as f:
            callback, find = make_callback(
                repo_root, os.path.dirname(build_file), target_rule_lookup
            )
            load = make_load_rules(repo_root, os.path.dirname(build_file))

            __builtins__["callback"] = callback
            __builtins__["find"] = find
            __builtins__["load"] = load

            frame = {
                **frame,
                "__builtins__": __builtins__,
            }

            reset_mock_imports(frame, ["callback", "find", "load"])

            exec(f.read(), frame)
    return target_rule_lookup



from __future__ import annotations

from dataclasses import dataclass, field
from queue import Queue
from threading import Lock
from typing import Dict, Optional, Set

from loader import Rule
from monitoring import StatusMonitor


@dataclass
class BuildState:
    # config parameters
    cache_directory: str
    target_rule_lookup: Dict[str, Rule]
    source_files: Set[str]
    repo_root: str

    # logging
    status_monitor: StatusMonitor = None

    # dynamic state
    scheduling_lock: Lock = field(default_factory=Lock)
    ready: Set[Rule] = field(default_factory=set)
    scheduled_but_not_ready: Set[Rule] = field(default_factory=set)
    work_queue: Queue[Optional[Rule]] = field(default_factory=Queue)



from __future__ import annotations

import hashlib
from pathlib import Path
from typing import Callable, Collection, Optional, Sequence

from cache import make_cache_memorize
from context import MemorizeContext
from fs_utils import copy_helper
from loader import Rule
from monitoring import log
from utils import BuildException, HashState, MissingDependency
from build_state import BuildState

from common.shell_utils import sh as run_shell


class ExecutionContext(MemorizeContext):
    def __init__(
        self,
        repo_root: str,
        cwd: str,
        hashstate: HashState,
        load_deps: Callable[[Sequence[str]], None],
        memorize: Callable[[str, str, str], None],
    ):
        super().__init__(repo_root, cwd, hashstate)
        self.load_deps = load_deps
        self.memorize = memorize

    def sh(self, cmd: str):
        super().sh(cmd)
        run_shell(cmd, shell=True, cwd=self.cwd, quiet=True)

    def add_deps(self, deps: Sequence[str]):
        super().add_deps(deps)
        self.load_deps([self.absolute(dep) for dep in deps])

    def input(self, *, file: str = None, sh: str = None):
        super().input(file=file, sh=sh)
        if file is not None:
            self.load_deps([self.absolute(file)])
            with open(self.absolute(file), "r") as f:
                return f.read()
        else:
            out = run_shell(
                sh, shell=True, cwd=self.cwd, capture_output=True, quiet=True
            ).decode("utf-8")
            self.memorize(
                self.hashstate.state(), hashlib.md5(sh.encode("utf-8")).hexdigest(), out
            )
            return out


def build(
    build_state: BuildState,
    rule: Rule,
    deps: Collection[str],
    *,
    scratch_path: Optional[Path],
):
    """
    All the dependencies that can be determined from caches have been
    obtained. Now we need to run. Either we will successfully finish everything,
    or we will get a missing dependency and have to requeue
    """
    memorize = make_cache_memorize(build_state.cache_directory)

    in_sandbox = scratch_path is not None

    loaded_deps = set()

    def load_deps(deps):
        deps = set(deps) - loaded_deps
        # check that these deps are built! Since they have not been checked by the PreviewExecution.
        missing_deps = []
        for dep in deps:
            if dep in build_state.target_rule_lookup:
                if build_state.target_rule_lookup[dep] not in build_state.ready:
                    missing_deps.append(dep)
            else:
                if dep not in build_state.source_files:
                    raise BuildException(f"Dependency missing: {dep}")
        if missing_deps:
            raise MissingDependency(*missing_deps)
        loaded_deps.update(deps)
        if in_sandbox:
            log(f"Loading dependencies {deps} into sandbox")
            copy_helper(
                src_root=build_state.repo_root,
                dest_root=scratch_path,
                src_names=deps,
                symlink=True,
            )

    load_deps(deps)
    hashstate = HashState()
    for dep in rule.deps:
        hashstate.update(dep.encode("utf-8"))
        with open(dep, "rb") as f:
            hashstate.update(f.read())

    ctx = ExecutionContext(
        scratch_path if in_sandbox else build_state.repo_root,
        scratch_path.joinpath(rule.location)
        if in_sandbox
        else Path(build_state.repo_root).joinpath(rule.location),
        hashstate,
        load_deps,
        memorize,
    )

    rule.impl(ctx)

    if in_sandbox:
        try:
            copy_helper(
                src_root=scratch_path,
                src_names=rule.outputs,
                dest_root=build_state.repo_root,
            )
        except FileNotFoundError as e:
            raise BuildException(
                f"Output file {e.filename} from rule {rule} was not generated."
            )

    for input_path in ctx.inputs:
        hashstate.update(input_path.encode("utf-8"))
        with open(input_path, "rb") as f:
            hashstate.update(f.read())

    return hashstate.state()
